@@ -5,6 +5,13 @@ use 5.006000;
 use strict;
 use warnings;
 use IO::Handle;
+#use IO::Compress::Deflate qw(deflate $DeflateError);
+#use IO::Uncompress::Inflate qw(inflate $InflateError);
+use Compress::Raw::Zlib;
+use Errno;
+use Socket qw( SOCK_DGRAM SOCK_STREAM SOCK_RAW PF_INET SOL_SOCKET IPPROTO_IP SO_ERROR IPPROTO_IP IP_TOS IP_TTL
+               inet_aton sockaddr_in  SO_SNDTIMEO SO_RCVTIMEO SO_KEEPALIVE);
+
 
 require Exporter;
 use AutoLoader qw(AUTOLOAD);
@@ -27,7 +34,7 @@ our @EXPORT = qw(
 
 		);
 
-our $VERSION = '0.61';
+our $VERSION = '0.84';
 # Preloaded methods go here.
 # Autoload methods go after =cut, and are processed by the autosplit program.
 use constant BUFSIZE => 4096;
@@ -50,10 +57,20 @@ sub new($$) {
 	$self->{"Encrypt"}=1;
 	$self->{"Debug"}=0;
 	$self->{"ErrMSG"}=undef;
+	#$self->{"ForcePASSVIP"}=undef;
 	$self->{"GetUpdateCallback"}  = undef;
 	$self->{"GetDoneCallback"}    = undef;
 	$self->{"PutUpdateCallback"}  = undef;
 	$self->{"PutDoneCallback"}    = undef;
+	$self->{"CompressionLevel"}=1;
+	$self->{"CompressionWindowBits"}=8;
+	$self->{"CompressionBufsize"}=8192;
+	$self->{"CompressionStrategy"}=4;
+#define Z_FILTERED            1
+#define Z_HUFFMAN_ONLY        2
+#define Z_RLE                 3
+#define Z_FIXED               4
+#define Z_DEFAULT_STRATEGY    0
 	return $self;
 };
 
@@ -84,6 +101,15 @@ sub rmdir ($$) {
 	$self->command("RMD $data");
 }
 
+sub bin ($) {
+	my ($self)=@_;
+	$self->command("TYPE I");
+}
+sub ascii ($) {
+	my ($self)=@_;
+	$self->command("TYPE A");
+}
+
 sub size ($$) {
 	my ($self,$filename)=@_;
 	my $size=$self->command("SIZE $filename");chop $size if defined($size);
@@ -130,8 +156,8 @@ sub readln {
         return $ln;
 };
 
-sub SOL_IP { 0; };
-sub IP_TOS { 1; };
+#sub SOL_IP { 0; };
+#sub IP_TOS { 1; };
 
 sub open($$$) {
 	my ($self,$host,$port)=@_;
@@ -142,14 +168,18 @@ sub open($$$) {
 	$self->{'Host'}=$host;
 	$self->{'Port'}=$port;
 	#tmp
-	use Socket;
+	
 	#setsockopt($sock,&SOL_SOCKET,&SO_KEEPALIVE,undef,undef) || warn "setsockopt: $!";
-	setsockopt($sock, SOL_SOCKET, SO_SNDTIMEO, pack('LL', 15, 0) ) or die "setsockopt".$!;  
-	setsockopt($sock, SOL_SOCKET, SO_RCVTIMEO, pack('LL', 15, 0) ) or die "setsockopt".$!;
-	setsockopt($sock, SOL_SOCKET, SO_KEEPALIVE, 1 ) or die "setsockopt".$!;  
+	setsockopt($sock, SOL_SOCKET, SO_SNDTIMEO, pack('L!L!', 15, 0) ) or warn "setsockopt SNDTIMEO unset: ".$!;  
+	setsockopt($sock, SOL_SOCKET, SO_RCVTIMEO, pack('L!L!', 15, 0) ) or warn "setsockopt RCVTIMEO unset: ".$!;
+	setsockopt($sock, SOL_SOCKET, SO_KEEPALIVE, 1 ) or warn "setsockopt KEEPALIVE unset: ".$!;  
 	#setsockopt($sock, SOL_SOCKET, IP_TOS, IPTOS_LOWDELAY ) or die "setsockopt".$!;  
 	#/usr/share/perl/5.8.4/Net/Ping.pm:      setsockopt($self->{"fh"}, SOL_IP, IP_TOS(), pack("I*", $self->{'tos'}))
-	setsockopt($sock, SOL_IP, IP_TOS(), pack("I*",0x10 ));#LOWLATENCY
+	#setsockopt($sock, SOL_IP, IP_TOS(), pack("I*",0x10 ));
+	#LOWLATENCY
+	setsockopt($sock, IPPROTO_IP, IP_TOS, pack("I*", 0x10))
+		or warn "error configuring tos LOW LATENCY to $self->{'fh'} $!";
+
 
 	#/usr/include/linux/ip.h:#define IPTOS_LOWDELAY          0x10
 	##define IPTOS_THROUGHPUT        0x08
@@ -179,8 +209,21 @@ sub open($$$) {
 			print STDERR "Received: $data" if $self->{Debug};
 		}
 	}
+
+	if ($self->{'Integrity'}) {
+		$self->command("INTEGRITY H");# TODO
+	};
+
+	if ($self->{'Compress'}) {
+		$self->command("MODE Z");# TODO
+	};
 	$self->{'RAWSock'}=$sock;
 
+	if ($self->{'Compress'}) {
+		# IO::Compress::Deflate
+		# IO::Uncompress::Inflate
+	};
+
 	if ($self->{'Encrypt'} && (! $self->{'FTPS'} )) {
 		#{tie(*S, "Net::SSLeay::Handle", $sock);$sock = \*S;};
 		# Unique glob?
@@ -221,7 +264,8 @@ sub command ($$){
 	my ($self,$data)=@_;
 	print STDERR "Sending: ",$data."\n" if $self->{Debug};
 	my $sock=$self->{'Sock'};
-	print $sock $data."\r\n";
+#	print $sock $data."\r\n";
+ 	syswrite( $sock ,$data."\r\n");
 	return $self->response();
 }
 
@@ -229,13 +273,24 @@ sub response ($) {
 	my ($self)=@_;
 	my $sock=$self->{'Sock'};
 	my ($read,$resp,$code,$cont);
-	$read=($resp=<$sock>);
+	my $start=time;
+	$read=($resp=<$sock>);#, redo if (Errno::EAGAIN && (!defined($read)));
+         #my $recvbytes=$sock->recv($read,1024);#, redo if Errno::EAGAIN;
+	
+	#TODO skip this, if RAWSock is disconnected... otherwise there's no sense in waiting for data..
+
+	while (!defined($read) && Errno::EAGAIN && (time-$start<190)) {
+		$read=($resp=<$sock>);
+		#, redo until timeout or sth
+	}
+
 	if (!defined($read)) {
-	warn "Damn! undefined response (err:$!) {H: ".$self->{'Host'}." P:".$self->{'Port'}."}\n";# unless defined($read);
-	$self->{'FTPCODE'}=undef;
-	$self->{'FTPMSG'}=undef;
-	$self->{'FTPRAWMSG'}=undef;
-	return undef;# unless defined($read);
+		warn "EAGAIN again 190s! $!" if Errno::EAGAIN;
+		warn "Damn! undefined response (err:$!) {H: ".$self->{'Host'}." P:".$self->{'Port'}."}\n";# unless defined($read);
+		$self->{'FTPCODE'}=undef;
+		$self->{'FTPMSG'}=undef;
+		$self->{'FTPRAWMSG'}=undef;
+		return undef;# unless defined($read);
 	};
 	return $self->responserest($read);
 }
@@ -317,12 +372,37 @@ sub nlst {
 				print STDERR "SSL for data connection enabled...\n" if $self->{Debug};
 			};
 			my $tmp;
-			while ($tmp=<$socket>) {
-#print STDERR "G: $q";
-#chop($tmp);chop($tmp);#\r\n -> remove.
-				$tmp=~s/\r\n$//;
-				push @files,$tmp;
-			};
+			if ( $self->{'Compress'} ) {
+				my $x = new Compress::Raw::Zlib::Inflate()
+					or die "Cannot create a inflation stream\n" ;
+				my $buf;my $read;
+
+				my ($output, $status) ;
+				while (sysread($socket,$buf,BUFSIZE)) 
+				{
+					$status = $x->inflate($buf, $output) ;
+					$read+=length($output);
+					#print $output ;
+ 					print STDERR "LST: $output\n" if $self->{Debug};
+					$output=~s/\r\n$//;
+					push @files,$output;
+
+					last if $status != Z_OK ;
+				};
+				die "inflation failed\n"
+					    unless $status == Z_STREAM_END ;
+
+#				print STDERR "===> READ: $read\n";
+
+			} else {
+			#TODO: if {'Compress'}
+				while ($tmp=<$socket>) {
+	#print STDERR "G: $q";
+	#chop($tmp);chop($tmp);#\r\n -> remove.
+					$tmp=~s/\r\n$//;
+					push @files,$tmp;
+				};
+			}
 		};
 		close $socket;
 		if ($response && ($self->{'FTPCODE'}<200) ) {if ($response) {$response=$self->response();};}
@@ -339,7 +419,7 @@ sub putblat {
 	$local=$remote unless defined($local);
 	$self->command("TYPE I") unless ($self->{'DontDoType'});
 	$socket=$self->datasocket();
-	die "SOCKET NOT CONNECTED! $!\n" unless defined($socket);
+	warn "SOCKET NOT CONNECTED! $!\n" unless defined($socket);
 	if ($self->{"EncryptData"}!=0) {$self->command("PROT P"); };
 	my $r=$self->command("$stororappe $remote");
 	if (!$r) {
@@ -356,20 +436,84 @@ sub putblat {
 	select($socket);
 #print "selected.\n";
 	if ($putorblat=~/put/) {
+		# 
+
+
 		CORE::open(L,"$local") or die "Can't open file $local, $!";
 		binmode L;
-		my $tmp;
-		while ($tmp=<L>) {
-			print $tmp;
-			if (defined ($self->{'PutUpdateCallback'})) {$self->{'PutUpdateCallback'}->( length($tmp) ); };#TODO send sth..
-		};#Probably syswrite/sysread would be smarter..
+		my $buf;
+
+		if ($self->{'Compress'}) {
+			#my $x = new Compress::Raw::Zlib::Deflate ( -Bufsize => 8192, -Level=> 1,-Strategy=>Z_DEFAULT_STRATEGY,-WindowBits=>15) 
+			my $x = new Compress::Raw::Zlib::Deflate ( -Bufsize => $self->{"CompressionBufsize"}, -Level=> $self->{"CompressionLevel"},-Strategy=>$self->{"CompressionStrategy"},-WindowBits=>$self->{"CompressionWindowBits"}) 
+				or die "Cannot create compression stream $!";
+			#deflate $local => $socket,AutoClose=>1,BinModeIn=>1
+		#		or die "Cannot compress to $socket $DeflateError $!";
+				#
+			my $read=0;my $circular=0;
+			my ($output, $status) ;
+			#while ($buf=<L>)
+			while ( sysread(L,$buf,BUFSIZE))
+			{
+				$status = $x->deflate($buf, $output) ;
+				$status == Z_OK
+					or die "deflation failed $!\n" ;
+				print $output ;
+				#print STDERR "GOT ".length($output)." bytes from deflation\n";
+				$read+=length($buf);
+				#$circular+=length($buf);
+				#if ($circular>8192) {
+				#	$circular=0;
+				#	print STDERR "trying part-flush\n";
+				#	$status = $x->flush($output,Z_SYNC_FLUSH) ;
+				#	$status == Z_OK
+				#		or die "deflation failed $!\n" ;
+				#	print $output ;
+				#	print STDERR "GOT ".length($output)." bytes from no-flush-deflation\n";
+				#};
+ 				if (defined ($self->{'PutUpdateCallback'})) {$self->{'PutUpdateCallback'}->( length($buf) ); };#TODO send sth..
+
+			}
+			print STDERR "out of buf LOOP\n" if $self->{Debug};
+			$status = $x->flush($output) ;
+			$status == Z_OK
+				or die "deflation failed $!\n" ;
+			#print $output ;
+			syswrite($socket, $output) ;
+			print STDERR "GOT ".length($output)." bytes from flush-deflation\n" if $self->{Debug};
+
+		} 
+		else {
+			while ($buf=<L>) {
+				print $buf;
+				if (defined ($self->{'PutUpdateCallback'})) {$self->{'PutUpdateCallback'}->( length($buf) ); };#TODO send sth..
+			};#Probably syswrite/sysread would be smarter..
+		}
 		close L;
 	} else {
-		print $local;
+		if ($self->{'Compress'}) {
+			my $x = new Compress::Raw::Zlib::Deflate ( -Bufsize => 8192) 
+				or die "Cannot create compression stream $!";
+			my $output;
+			my $status = $x->deflate($local, $output) ;
+			$status == Z_OK
+				or die "deflation failed\n" ;
+			#print $output ;
+			syswrite($socket, $output) ;
+			$status = $x->flush($output) ;
+			$status == Z_OK
+				or die "deflation failed\n" ;
+			#print $output ;
+			syswrite($socket, $output) ;
+		} else {
+			#print $local;
+			syswrite($socket, $local) ;
+		};
 		if (defined ($self->{'PutUpdateCallback'})) {$self->{'PutUpdateCallback'}->( length($local) ); };#TODO send sth..
 	}
 #print "after write...\n";
 	select(STDOUT);
+	$socket->flush();
 	close $socket;
 	my $response=$self->response();
 	print  STDERR "resp(after$stororappe) ",$response if $self->{Debug};
@@ -404,12 +548,11 @@ sub getslurp {
 	$self->command("TYPE I");
 	$socket=$self->datasocket();
 	#tmp
-	use Socket;
 	#setsockopt($sock,&SOL_SOCKET,&SO_KEEPALIVE,undef,undef) || warn "setsockopt: $!";
-	setsockopt($socket, SOL_SOCKET, SO_SNDTIMEO, pack('LL', 15, 0) ) or die "setsockopt".$!;  
-	setsockopt($socket, SOL_SOCKET, SO_RCVTIMEO, pack('LL', 15, 0) ) or die "setsockopt".$!;
-	setsockopt($socket, SOL_SOCKET, SO_KEEPALIVE, 1 ) or die "setsockopt".$!;  
-	setsockopt($sock, SOL_IP, IP_TOS(), pack("I*",0x08 ));#THROUGHPUT
+	setsockopt($socket, SOL_SOCKET, SO_SNDTIMEO, pack('L!L!', 15, 0) ) or warn "setsockopt SNDTIMEO: ".$!;  
+	setsockopt($socket, SOL_SOCKET, SO_RCVTIMEO, pack('L!L!', 15, 0) ) or warn "setsockopt RCVTIMEO: ".$!;
+	setsockopt($socket, SOL_SOCKET, SO_KEEPALIVE, 1 ) or warn "setsockopt KEEPALIVE: ".$!;  
+	setsockopt($socket, IPPROTO_IP, IP_TOS(), pack("I*",0x08 )) or warn "setsockopt TOS THROUGPHUT problem: $!";#THROUGHPUT
 	#end tmp 2008-11-04
 
 	if ($self->{"EncryptData"}!=0) {$self->command("PROT P"); };
@@ -428,27 +571,83 @@ sub getslurp {
 		print STDERR "getorslurp: get\n" if $self->{Debug};
 		CORE::open(L,">$local") or die("Can't open file for writing $local, $!");
 		binmode L;
-		# TODO replace while <$socket> with
-		# TODO while sysread($sock,$tmp,BUFSIZE);
-		my $tmp;
-		while (sysread($socket,$tmp,BUFSIZE)) {
-			print L $tmp;
-			print STDERR length($tmp),":;\n" if $self->{Debug};
-			if (defined ($self->{'GetUpdateCallback'})) {$self->{'GetUpdateCallback'}->(length($tmp)); };#TODO send sth..
-		};
-#tmp-2008-10-22#		while ($tmp=<$socket>) {
-#tmp-2008-10-22#			print L $tmp;
-#tmp-2008-10-22#			print STDERR length($tmp),":;\n" if $self->{Debug};
-#tmp-2008-10-22#			if (defined ($self->{'GetUpdateCallback'})) {$self->{'GetUpdateCallback'}->();print STDERR "GUC defined, and has been called\n"; };#TODO send sth..
-#tmp-2008-10-22#		};
-		close L;
+		my $tmp;my $cntr=0;
+
+
+		if ($self->{'Compress'}) {
+			#my $z = new IO::Uncompress::Inflate $socket, AutoClose=>1,Strict=>1
+			#	or die "IO::Uncompress::Inflate failed: $InflateError\n";
+			#inflate $socket => "ftp.".$local, AutoClose=>1 
+			#	or die "Cannot decompress from $socket to $local $InflateError $!";
+
+			        
+			my $x = new Compress::Raw::Zlib::Inflate()
+				or die "Cannot create a inflation stream\n" ;
+			my $buf;my $read;
+
+			my ($output, $status) ;
+			while (sysread($socket,$buf,BUFSIZE)) 
+			{
+				$status = $x->inflate($buf, $output) ;
+				$read+=length($output);
+				#print $output ;
+				print L $output;
+				print STDERR length($output),":;".++$cntr."\n" if $self->{Debug};
+				if (defined ($self->{'GetUpdateCallback'})) {$self->{'GetUpdateCallback'}->(length($output)); };#TODO send sth..
+				last if $status != Z_OK ;
+			};
+			die "inflation failed\n"
+				    unless $status == Z_STREAM_END ;
+
+#			print STDERR "===> READ: $read\n";
+		} else 
+		{
+# TODO replace while <$socket> with
+# TODO while sysread($sock,$tmp,BUFSIZE);
+
+			while (sysread($socket,$tmp,BUFSIZE)) {
+				print L $tmp;
+				print STDERR length($tmp),":;".++$cntr."\n" if $self->{Debug};
+				if (defined ($self->{'GetUpdateCallback'})) {$self->{'GetUpdateCallback'}->(length($tmp)); };#TODO send sth..
+			};
+	#tmp-2008-10-22#		while ($tmp=<$socket>) {
+	#tmp-2008-10-22#			print L $tmp;
+	#tmp-2008-10-22#			print STDERR length($tmp),":;\n" if $self->{Debug};
+	#tmp-2008-10-22#			if (defined ($self->{'GetUpdateCallback'})) {$self->{'GetUpdateCallback'}->();print STDERR "GUC defined, and has been called\n"; };#TODO send sth..
+	#tmp-2008-10-22#		};
+			close L;
+		}
 	} else {
+
 		print STDERR "getorslurp: slurp($getorslurp)\n" if $self->{Debug};
 		my $tmp;
-		while ($tmp=<$socket>) {
-            $slurped.=$tmp;print STDERR ":." if $self->{Debug}; 
-            if (defined ($self->{'GetUpdateCallback'})) {$self->{'GetUpdateCallback'}->( length($tmp) ); };#TODO send sth..
-        };
+
+		if ($self->{'Compress'}) {
+			my $x = new Compress::Raw::Zlib::Inflate()
+				or die "Cannot create a inflation stream\n" ;
+			my $buf;my $read;
+
+			my ($output, $status) ;
+			while (sysread($socket,$buf,BUFSIZE)) 
+			{
+				$status = $x->inflate($buf, $output) ;
+				$read+=length($output);
+				#print $output ;
+				$slurped.=$output;
+				print STDERR length($output),":slurpZ;" if $self->{Debug};
+				if (defined ($self->{'GetUpdateCallback'})) {$self->{'GetUpdateCallback'}->(length($output)); };#TODO send sth..
+				last if $status != Z_OK ;
+			};
+			die "inflation failed\n"
+				    unless $status == Z_STREAM_END ;
+
+
+		} else {
+			while ($tmp=<$socket>) {
+				$slurped.=$tmp;print STDERR ":slurp." if $self->{Debug}; 
+				if (defined ($self->{'GetUpdateCallback'})) {$self->{'GetUpdateCallback'}->( length($tmp) ); };#TODO send sth..
+			};
+		};
 	};
 	close $socket;
 	my $response=$self->response();
@@ -464,7 +663,10 @@ sub datasocket {
 		if ($self->msgcode()==227 &&  $tmp=~/[^\d]*(\d+),(\d+),(\d+),(\d+),(\d+),(\d+)/) {
 			my $port=$5*256+$6;
 			my $host="$1.$2.$3.$4";
+			print STDERR "PASV port is $port ( $host )\n" if $self->{Debug};
+			if (defined($self->{'ForcePASVIP'})) {$host=$self->{'ForcePASVIP'};print STDERR "Forcing PASV IP to $host\n" if $self->{Debug}};
 			$socket = Net::SSLeay::Handle->make_socket($host, $port);
+			setsockopt($socket, IPPROTO_IP, IP_TOS(), pack("I*",0x08 ));#THROUGHPUT
 			if (defined($socket)) {
 				print STDERR "Data link connected.. to $host at $port\n" if $self->{Debug};
 			} else {
@@ -518,6 +720,16 @@ sub setup_protection {
 };
 
 
+sub check_raw_socket {
+	my ($self)=@_;
+	use Socket;
+	my $s=$self->{'RAWSock'};
+	my $hersockaddr    = getpeername($s);
+	my ($port, $iaddr) = sockaddr_in($hersockaddr);
+	return $s->eof."$hersockaddr $port $iaddr";
+};
+
+
 
 1;
 __END__
